#!/usr/bin/perl
=head1 NAME

queue-fast.pl - display/delete queue of qmail

=head1 SYNOPSIS

    perl queue-fast.pl [-d] [ADDRESSES...]

=head1 DESCRIPTION

    # display all mail in queue
    perl queue-fast.pl

    # display all mail to user@domain.name
    perl queue-fast.pl user@domain.name

    # delete all mail to user@domain.name
    perl queue-fast.pl -d user@domain.name

=head1 SEEL ALSO

http://www.kawa.net/works/qmail/queue-fast.html

=head1 COPYRIGHT

(C) Copyright 2004-2005 Kawasaki Yusuke

=cut

use strict;

my $QUEUE = "/var/qmail/queue";
my $lnum = [];

my @mode = grep {/^-/} @ARGV;
my @addr = grep {! /^-/} @ARGV;
my $match = { map {$_=>1} @addr } if scalar @addr;

my $delmode = undef;

if ( scalar @addr ) {
    print STDERR "# address: ", join( " ", @addr ), "\n";
}

foreach my $arg ( @mode ) {
    if ( $arg eq "-d" ) {
        die "delete mode but no address.\n" unless scalar @addr;
        $delmode ++;
        print STDERR "# delete mode enabled.\n";
    } else {
        die "arg '$arg' is not supported.\n";
    }
}


opendir( NUM, "$QUEUE/mess" ) or die "$! - $QUEUE/mess\n";
while ( 1 ) {
    my $num = readdir( NUM );
    last unless defined $num;
    next unless ( $num =~ /^\d+$/ );
    next unless ( -d "$QUEUE/mess/$num" );
    # print STDERR "$QUEUE/mess/$num\n";
    push( @$lnum, $num );
}
closedir( NUM );

@$lnum = sort { $a <=> $b } @$lnum;

my $cntnum = 0;
foreach my $num ( @$lnum ) {
    my $lmess = [];
    opendir( MESS, "$QUEUE/mess/$num" ) or die "$! - $QUEUE/mess/$num\n";
    while ( 1 ) {
        my $inode = readdir( MESS );
        last unless defined $inode;
        next unless ( $inode =~ /^\d+$/ );
        push( @$lmess, $inode );
    }
    close( MESS );
    printf STDERR ( "# %d messages found in %s/mess/%d (%d/%d)\n",
        scalar( @$lmess ), $QUEUE, $num, ++ $cntnum, scalar @$lnum );

    foreach my $inode ( @$lmess ) {
        my $pidfn = "$QUEUE/pid/$inode";
        my $messfn = "$QUEUE/mess/$num/$inode";
        my $intdfn = "$QUEUE/intd/$inode";
        my $todofn = "$QUEUE/todo/$inode";
        my $infofn = "$QUEUE/info/$num/$inode";
        my $local = "$QUEUE/local/$num/$inode";
        my $remote = "$QUEUE/remote/$num/$inode";
        my $bounce = "$QUEUE/bounce/$num/$inode";
        my @file = ( $messfn, $pidfn, $intdfn, $todofn,
                     $infofn, $local, $remote, $bounce );
        local $/ = "\0";
        my @from = ();
        my @rcpt = ();
        my $mode = "-";
        if ( open( INFO, $infofn )) {
            @from = <INFO>;
            close( INFO );
            if ( open( REMOTE, $remote )) {
                @rcpt = <REMOTE>;
                close( REMOTE );
                $mode = "R";
            } elsif ( open( LOCAL, $local )) {
                @rcpt = <LOCAL>;
                close( LOCAL );
                $mode = "L";
            }
        } elsif ( open( INTD, $intdfn )) {
            my @lines = <INTD>;
            close( INTD );
            @from = grep { /^F/ } @lines;
            @rcpt = grep { /^T/ } @lines;
            $mode = "Q";
        } else {
            $mode = "X";
            # print STDERR "$messfn is broken queue.\n";
            # next;
        }

        @from = grep {$_ ne ""} map {(/^F(.*)\0$/)[0]} @from;
        @rcpt = grep {$_ ne ""} map {(/^T(.*)\0$/)[0]} @rcpt;

        foreach ( @from, @rcpt ) {
            s/([\000-\040\%\177-\377])/sprintf("%%%02X",ord($1))/ges;
        }

        if ( ref $match ) {
            next if ( scalar @rcpt != 1 );
            next unless exists $match->{$rcpt[0]};
        }

        printf( "%s %s %s => %s\n", "$num/$inode", $mode, 
            (join(" ",@from)||"-"), (join(" ",@rcpt)||"-"));

        if ( $delmode ) {
            print "deleteing: ", join( " ", grep {-r $_} @file ), "\n";
            unlink( @file );
        }
    }
}

